import json
from calendar import monthrange
from datetime import timedelta, date, datetime

from django.conf import settings
from django.db.models import Q
from django.forms import CharField, ChoiceField
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.template import Template, Context, TemplateSyntaxError
from django.urls import reverse
from django.views.generic import TemplateView
from django_datatables.columns import ColumnBase, MenuColumn
from django_datatables.datatables import DatatableExcludedRow
from django_datatables.helpers import DUMMY_ID, row_link
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import MenuItem, HtmlMenu
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.helper import modal_button, modal_button_method
from django_modals.modals import ModelFormModal, Modal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.select2 import Select2Multiple, Select2

from advanced_report_builder.columns import ReportBuilderNumberColumn
from advanced_report_builder.data_merge.utils import get_menu_fields, get_data_merge_columns
from advanced_report_builder.data_merge.widget import DataMergeWidget
from advanced_report_builder.filter_query import FilterQueryMixin
from advanced_report_builder.globals import DATE_FORMAT_TYPES_DJANGO_FORMAT, DATE_FORMAT_TYPE_DD_MM_YY_SLASH, \
    DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY
from advanced_report_builder.models import KanbanReport, KanbanReportLane, ReportType, KanbanReportDescription
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import crispy_modal_link_args, get_field_details, get_report_builder_class
from advanced_report_builder.variable_date import VariableDate
from advanced_report_builder.views.charts_base import ChartJSTable
from advanced_report_builder.views.modals_base import QueryBuilderModalBase
from advanced_report_builder.views.report import ReportBase


class DescriptionColumn(ColumnBase):
    def row_result(self, data, _page_data, columns):
        html = self.options['html']

        for column in columns:
            if not isinstance(column, DescriptionColumn):
                data[column.column_name] = column.row_result(data, _page_data)

        try:
            template = Template(html)
            context = Context(data)
            return template.render(context)
        except TemplateSyntaxError as e:
            return f'Error in description ({e})'


class KanbanTable(ChartJSTable):
    def get_table_array(self, request, results):
        result_processes = self.get_result_processes()
        for p in result_processes:
            p.setup_results(request, self.page_results)
        for c in self.columns:
            c.setup_results(request, self.page_results)
        results_list = []
        for data_dict in results:
            try:
                for p in result_processes:
                    p.row_result(data_dict, self.page_results)
                row_data = []
                for column in self.columns:
                    if isinstance(column, DescriptionColumn):
                        row_data.append(column.row_result(data_dict, self.page_results, columns=self.columns))
                    else:
                        row_data.append(column.row_result(data_dict, self.page_results))

                results_list.append(row_data)
            except DatatableExcludedRow:
                pass
        return results_list


class KanbanView(ReportBase, FilterQueryMixin, TemplateView):
    number_field = ReportBuilderNumberColumn
    template_name = 'advanced_report_builder/kanban/report.html'
    chart_js_table = KanbanTable

    def __init__(self, *args, **kwargs):
        self.chart_report = None
        super().__init__(*args, **kwargs)

    def setup_menu(self):
        super().setup_menu()
        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()
        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(),
        )

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.kanbanreport
        self.dashboard_report = kwargs.get('dashboard_report')
        self.enable_edit = kwargs.get('enable_edit')
        return super().dispatch(request, *args, **kwargs)

    def view_filter_extra(self, query, table):
        if table.extra_query_filter:
            query = query.filter(table.extra_query_filter)
        return self.view_filter(query, table)

    def get_lane(self, base_model, kanban_report_lane, lanes, label=None, extra_query_filter=None, multiple=False):

        table = self.chart_js_table(model=base_model)

        report_builder_class = get_report_builder_class(model=base_model,
                                                        report_type=kanban_report_lane.report_type)
        table_indexes = []

        if kanban_report_lane.heading_field is not None:
            table_indexes.append('heading')
            table.add_columns(kanban_report_lane.heading_field)

        if kanban_report_lane.background_colour_field is not None:
            table_indexes.append('background_colour')
            table.add_columns(kanban_report_lane.background_colour_field)

        if kanban_report_lane.heading_colour_field is not None:
            table_indexes.append('heading_colour')
            table.add_columns(kanban_report_lane.heading_colour_field)

        if kanban_report_lane.kanban_report_description is not None:
            description = kanban_report_lane.kanban_report_description.description
            columns, column_map = get_data_merge_columns(base_model=base_model,
                                                         report_builder_class=report_builder_class,
                                                         html=description,
                                                         table=table)

            table_indexes.append('description')
            table.add_columns(DescriptionColumn(column_name='description',
                                                field='',
                                                html=description,
                                                column_map=column_map))
            table.add_columns(*columns)

        table.table_options['indexes'] = table_indexes

        if kanban_report_lane.link_field and self.kwargs.get('enable_links'):
            _, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                           field=kanban_report_lane.link_field,
                                                           report_builder_class=report_builder_class)
            if isinstance(col_type_override.field, list):
                field = col_type_override.field[0]
            else:
                field = 'id'
            if field not in table.columns:
                table.add_columns(f'.{field}')
            table.has_link = True
            table.table_options['row_href'] = row_link(col_type_override.url, field)
        else:
            table.has_link = False

        table.datatable_template = 'advanced_report_builder/kanban/middle.html'

        if kanban_report_lane.order_by_field:
            if kanban_report_lane.order_by_ascending:
                table.order_by = [kanban_report_lane.order_by_field]
            else:
                table.order_by = [f'-{kanban_report_lane.order_by_field}']

        table.query_data = kanban_report_lane.query_data
        table.extra_query_filter = extra_query_filter
        table.view_filter = self.view_filter_extra
        lanes.append({'datatable': table,
                      'label': label,
                      'kanban_report_lane': kanban_report_lane,
                      'multiple': multiple})

    @staticmethod
    def get_multiple_date(multiple_type, current_date):
        if multiple_type in (KanbanReportLane.MULTIPLE_TYPE_DAILY, KanbanReportLane.MULTIPLE_TYPE_DAILY_WITHIN):
            return current_date + timedelta(days=1)
        elif multiple_type in (KanbanReportLane.MULTIPLE_TYPE_WEEKLY, KanbanReportLane.MULTIPLE_TYPE_WEEKLY_WITHIN):
            return current_date + timedelta(days=7)
        elif multiple_type in (KanbanReportLane.MULTIPLE_TYPE_MONTHLY, KanbanReportLane.MULTIPLE_TYPE_MONTHLY_WITHIN):
            start_date = date(current_date.year, current_date.month, 1)
            number_of_days = monthrange(start_date.year, current_date.month)[1]
            result = datetime.combine(start_date, datetime.min.time()) + timedelta(days=number_of_days)
            return result

        assert False

    @staticmethod
    def get_date_format():
        return DATE_FORMAT_TYPES_DJANGO_FORMAT[DATE_FORMAT_TYPE_DD_MM_YY_SLASH]

    @staticmethod
    def get_month_format():
        return DATE_FORMAT_TYPES_DJANGO_FORMAT[DATE_FORMAT_TYPE_SHORT_WORDS_MM_YY]

    def get_full_label(self, multiple_type, current_date, label):
        if multiple_type in (KanbanReportLane.MULTIPLE_TYPE_DAILY, KanbanReportLane.MULTIPLE_TYPE_DAILY_WITHIN):
            date_string = date.strftime(current_date, self.get_date_format())
            return f'{label} {date_string}'
        elif multiple_type in (KanbanReportLane.MULTIPLE_TYPE_WEEKLY, KanbanReportLane.MULTIPLE_TYPE_WEEKLY_WITHIN):
            date_string = date.strftime(current_date, self.get_date_format())
            week_number = current_date.isocalendar()[1]
            return f'{label} {date_string} (w/c {week_number})'
        elif multiple_type in (KanbanReportLane.MULTIPLE_TYPE_MONTHLY, KanbanReportLane.MULTIPLE_TYPE_MONTHLY_WITHIN):
            date_string = date.strftime(current_date, self.get_month_format())
            return f'{label} {date_string}'
        assert False

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = self.get_title()
        kanban_report_lanes = self.chart_report.kanbanreportlane_set.all()
        lanes = []
        headings = []

        for kanban_report_lane in kanban_report_lanes:
            base_model = kanban_report_lane.get_base_modal()
            report_builder_class = get_report_builder_class(model=base_model,
                                                            report_type=kanban_report_lane.report_type)
            if kanban_report_lane.multiple_type == KanbanReportLane.MULTIPLE_TYPE_NA:
                self.get_lane(base_model=base_model,
                              kanban_report_lane=kanban_report_lane,
                              lanes=lanes)
                headings.append({'label': kanban_report_lane.name,
                                 'row_span': 2,
                                 'col_span': 1})
            else:
                variable_date = VariableDate()
                start_date_and_time, _, _ = variable_date.get_variable_dates(kanban_report_lane.multiple_start_period)
                _, end_date_and_time, _ = variable_date.get_variable_dates(kanban_report_lane.multiple_end_period)

                multiple_type = kanban_report_lane.multiple_type
                current_start_date = start_date_and_time

                _, _, _, field_name = get_field_details(base_model=base_model,
                                                        field=kanban_report_lane.multiple_type_date_field,
                                                        report_builder_class=report_builder_class)

                end_field_name = None
                if kanban_report_lane.multiple_type_end_date_field:
                    _, _, _, end_field_name = get_field_details(base_model=base_model,
                                                                field=kanban_report_lane.multiple_type_end_date_field,
                                                                report_builder_class=report_builder_class)

                sub_lanes = []
                while (current_end_date := self.get_multiple_date(
                        multiple_type=multiple_type,
                        current_date=current_start_date)) <= end_date_and_time:

                    if multiple_type in (KanbanReportLane.MULTIPLE_TYPE_DAILY,
                                         KanbanReportLane.MULTIPLE_TYPE_WEEKLY,
                                         KanbanReportLane.MULTIPLE_TYPE_MONTHLY):
                        query_filter = Q((field_name + "__gte", current_start_date)) &\
                                       (Q((field_name + "__lt", current_end_date)))
                    else:

                        query_filter = Q((end_field_name + "__gte", current_start_date)) &\
                                        (Q((field_name + "__lt", current_end_date)))

                    label = self.get_full_label(multiple_type=multiple_type,
                                                current_date=current_start_date,
                                                label=kanban_report_lane.multiple_type_label)

                    self.get_lane(base_model=base_model,
                                  kanban_report_lane=kanban_report_lane,
                                  lanes=sub_lanes,
                                  label=label,
                                  extra_query_filter=query_filter,
                                  multiple=True)
                    current_start_date = current_end_date

                headings.append({'label': kanban_report_lane.name,
                                 'col_span': len(sub_lanes),
                                 'row_span': 1})

                lanes += sub_lanes

        context['kanban_report'] = self.chart_report
        context['headings'] = headings
        context['lanes'] = lanes
        return context

    def view_filter(self, query, table):
        if not table.query_data:
            return query

        return self.process_query_filters(query=query,
                                          search_filter_data=table.query_data)

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    def pod_report_menu(self):
        return [MenuItem(f'advanced_report_builder:kanban_modal,pk-{self.chart_report.id}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary']),
                *self.duplicate_menu(request=self.request, report_id=self.chart_report.id)
                ]

    def pod_dashboard_edit_menu(self):
        return [MenuItem(f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                         menu_display='Edit',
                         font_awesome='fas fa-pencil-alt', css_classes=['btn-primary'])]

    # noinspection PyMethodMayBeStatic
    def queries_menu(self):
        return []


class KanbanModal(ModelFormModal):
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = KanbanReport
    widgets = {'report_tags': Select2Multiple}
    ajax_commands = ['datatable', 'button']

    form_fields = ['name',
                   'notes',
                   'report_tags']

    def form_setup(self, form, *_args, **_kwargs):
        org_id = self.object.id if hasattr(self, 'object') else None
        form.fields['notes'].widget.attrs['rows'] = 3
        if org_id is not None:
            lane_menu_items = [MenuItem(f'advanced_report_builder:kanban_lane_duplicate_modal,pk-{DUMMY_ID}',
                                        menu_display='Duplicate',
                                        css_classes='btn btn-sm btn-outline-dark',
                                        font_awesome='fas fa-clone'),
                               MenuItem(f'advanced_report_builder:kanban_lane_modal,pk-{DUMMY_ID}',
                                        menu_display='Edit',
                                        css_classes='btn btn-sm btn-outline-dark',
                                        font_awesome='fas fa-pencil')]

            form.fields['lanes'] = CharField(
                required=False,
                label='Lanes',
                widget=DataTableReorderWidget(
                    model=KanbanReportLane,
                    order_field='order',
                    fields=['_.index',
                            '.id',
                            'name',
                            MenuColumn(column_name='menu', field='id',
                                       column_defs={'orderable': False, 'className': 'dt-right'},
                                       menu=HtmlMenu(self.request, 'button_group').add_items(*lane_menu_items)),
                            ],
                    attrs={'filter': {'kanban_report__id': self.object.id}})
            )

            description_menu_items = [
                MenuItem(f'advanced_report_builder:kanban_description_duplicate_modal,pk-{DUMMY_ID}',
                         menu_display='Duplicate',
                         css_classes='btn btn-sm btn-outline-dark',
                         font_awesome='fas fa-clone'),
                MenuItem(f'advanced_report_builder:kanban_description_modal,pk-{DUMMY_ID}',
                         menu_display='Edit',
                         css_classes='btn btn-sm btn-outline-dark',
                         font_awesome='fas fa-pencil')]

            form.fields['descriptions'] = CharField(
                required=False,
                label='Descriptions',
                widget=DataTableReorderWidget(
                    model=KanbanReportDescription,
                    order_field='order',
                    fields=['_.index',
                            '.id',
                            'name',
                            MenuColumn(column_name='menu', field='id',
                                       column_defs={'orderable': False, 'className': 'dt-right'},
                                       menu=HtmlMenu(self.request, 'button_group').add_items(*description_menu_items)),
                            ],
                    attrs={'filter': {'kanban_report__id': self.object.id}})
            )

            modal_button('Custom Close', 'close', 'btn-warning')

            return [*self.form_fields,
                    crispy_modal_link_args('advanced_report_builder:kanban_lane_modal', 'Add Lane',
                                           'kanban_report_id-', self.object.id, div=True,
                                           div_classes='form-buttons', button_classes='btn btn-primary',
                                           font_awesome='fa fa-plus'),
                    'lanes',
                    crispy_modal_link_args('advanced_report_builder:kanban_description_modal', 'Add Description',
                                           'kanban_report_id-', self.object.id, div=True,
                                           div_classes='form-buttons', button_classes='btn btn-primary',
                                           font_awesome='fa fa-plus'),
                    'descriptions',
                    ]

    def datatable_sort(self, **kwargs):

        form = self.get_form()
        widget = form.fields[kwargs['table_id'][3:]].widget
        _model = widget.attrs['table_model']
        current_sort = dict(_model.objects.filter(kanban_report=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = _model.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('')

    def post_save(self, created, form):
        if created:
            self.modal_redirect(self.request.resolver_match.view_name, slug=f'pk-{self.object.id}-new-True')
        else:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if url_name and self.slug.get('new'):
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.command_response('redirect', url=url)


class KanbanLaneModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/kanban/modal.html'
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = KanbanReportLane
    helper_class = HorizontalNoEnterHelper

    widgets = {'report_tags': Select2Multiple,
               'order_by_ascending': RBToggle}

    form_fields = ['name',
                   'report_type',
                   'heading_field',
                   'kanban_report_description',
                   'order_by_field',
                   'order_by_ascending',
                   'multiple_type',
                   'multiple_type_label',
                   'multiple_type_date_field',
                   'multiple_type_end_date_field',
                   'multiple_start_period',
                   'multiple_end_period',
                   'link_field',
                   'background_colour_field',
                   'heading_colour_field',
                   'query_data']

    def form_setup(self, form, *_args, **_kwargs):
        form.add_trigger('multiple_type', 'onchange', [
            {'selector': '#div_id_multiple_type_label,#div_id_multiple_type_date_field,'
                         '#div_id_multiple_start_period,#div_id_multiple_end_period',
             'values': {KanbanReportLane.MULTIPLE_TYPE_NA: 'hide'}, 'default': 'show'},
            {'selector': '#div_id_multiple_type_end_date_field',
             'values': {KanbanReportLane.MULTIPLE_TYPE_DAILY_WITHIN: 'show',
                        KanbanReportLane.MULTIPLE_TYPE_WEEKLY_WITHIN: 'show',
                        KanbanReportLane.MULTIPLE_TYPE_MONTHLY_WITHIN: 'show'},
             'default': 'hide'},
        ])

        if 'data' in _kwargs:
            heading_field = _kwargs['data'].get('heading_field')
            order_by_field = _kwargs['data'].get('order_by_field')
            multiple_type_date_field = _kwargs['data'].get('multiple_type_date_field')
            multiple_type_end_date_field = _kwargs['data'].get('multiple_type_end_date_field')
            background_colour_field = _kwargs['data'].get('background_colour_field')
            heading_colour_field = _kwargs['data'].get('heading_colour_field')
            link_field = _kwargs['data'].get('link_field')
            report_type_id = _kwargs['data'].get('report_type')
            kanban_report_description = KanbanReportDescription.objects.filter(
                pk=_kwargs['data'].get('kanban_report_description')).first()
            report_type = get_object_or_404(ReportType, id=report_type_id)

        else:
            heading_field = form.instance.heading_field
            order_by_field = form.instance.order_by_field
            report_type = form.instance.report_type
            multiple_type_date_field = form.instance.multiple_type_date_field
            multiple_type_end_date_field = form.instance.multiple_type_end_date_field
            background_colour_field = form.instance.background_colour_field
            heading_colour_field = form.instance.heading_colour_field
            kanban_report_description = form.instance.kanban_report_description
            link_field = form.instance.link_field

        form.fields['kanban_report_description'].widget = Select2(attrs={'ajax': True})
        if kanban_report_description is not None:
            form.fields['kanban_report_description'].widget.select_data = [{'id': kanban_report_description.id,
                                                                            'text': kanban_report_description.name}]
        form.fields['kanban_report_description'].label = 'Description'

        self.setup_field(field_type='all',
                         form=form,
                         field_name='heading_field',
                         selected_field_id=heading_field,
                         report_type=report_type)

        self.setup_field(field_type='all',
                         form=form,
                         field_name='order_by_field',
                         selected_field_id=order_by_field,
                         report_type=report_type)

        self.setup_field(field_type='date',
                         form=form,
                         field_name='multiple_type_date_field',
                         selected_field_id=multiple_type_date_field,
                         report_type=report_type)

        self.setup_field(field_type='date',
                         form=form,
                         field_name='multiple_type_end_date_field',
                         selected_field_id=multiple_type_end_date_field,
                         report_type=report_type)

        self.setup_field(field_type='link',
                         form=form,
                         field_name='link_field',
                         selected_field_id=link_field,
                         report_type=report_type)

        self.setup_field(field_type='colour',
                         form=form,
                         field_name='background_colour_field',
                         selected_field_id=background_colour_field,
                         report_type=report_type)

        self.setup_field(field_type='colour',
                         form=form,
                         field_name='heading_colour_field',
                         selected_field_id=heading_colour_field,
                         report_type=report_type)

        range_type_choices = VariableDate.RANGE_TYPE_CHOICES
        form.fields['multiple_start_period'] = ChoiceField(required=False, choices=range_type_choices)
        form.fields['multiple_end_period'] = ChoiceField(required=False, choices=range_type_choices)

        return ('name',
                'report_type',
                'heading_field',
                'kanban_report_description',
                'order_by_field',
                'order_by_ascending',
                'multiple_type',
                'multiple_type_label',
                'multiple_type_date_field',
                'multiple_type_end_date_field',
                'multiple_start_period',
                'multiple_end_period',
                'background_colour_field',
                'heading_colour_field',
                'link_field',
                FieldEx('query_data',
                        template='advanced_report_builder/query_builder.html'),
                )

    def select2_link_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='link',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_heading_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='all',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_order_by_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='all',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_multiple_type_date_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='date',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_multiple_type_end_date_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='date',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_background_colour_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='colour',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_heading_colour_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='colour',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_kanban_report_description(self, **kwargs):
        descriptions = []
        report_type_id = kwargs['report_type']
        if report_type_id != '':
            kanban_report_id = self.object.kanban_report_id

            _filter = KanbanReportDescription.objects.filter(report_type_id=report_type_id,
                                                             kanban_report_id=kanban_report_id)

            search = kwargs.get('search')
            if search:
                _filter = _filter.filter(name__icontains=search)

            for description in _filter.values('id', 'name'):
                descriptions.append({'id': description['id'], 'text': description['name']})

        return JsonResponse({'results': descriptions})

    def form_valid(self, form):
        form.save()
        return self.command_response('reload')


class KanbanLaneDuplicateModal(Modal):

    def modal_content(self):
        return 'Are you sure you want to duplicate this lane?'

    def get_modal_buttons(self):
        return [modal_button_method('Confirm', 'duplicate'), modal_button('Cancel', 'close', 'btn-secondary')]

    def button_duplicate(self, **_kwargs):
        kanban_report_lane = get_object_or_404(KanbanReportLane, id=self.slug['pk'])
        kanban_report_lane.pk = None
        kanban_report_lane.name = f'Copy of {kanban_report_lane.name}'
        kanban_report_lane.order = None
        kanban_report_lane.save()
        return self.command_response('reload')


class KanbanDescriptionModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/kanban/description_modal.html'
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = KanbanReportDescription

    widgets = {'report_tags': Select2Multiple,
               'order_by_ascending': RBToggle}

    form_fields = ['name',
                   'report_type',
                   'description']

    def form_setup(self, form, *_args, **_kwargs):
        form.fields['description'].widget = DataMergeWidget()

        return ('name',
                'report_type',
                'description')

    def ajax_get_description_data_merge_menu(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']
        menus = []
        if kwargs['report_type'] != '':
            report_builder_class, base_model = self.get_report_builder_class(report_type_id=kwargs['report_type'])

            get_menu_fields(base_model=base_model,
                            report_builder_class=report_builder_class,
                            menus=menus)

            menus = [{'code': 'data_merge',
                      'text': 'Data Merge',
                      'menu': menus}]

        return self.command_response(f'build_data_merge_menu_{field_auto_id}',
                                     data=json.dumps(menus))

    def form_valid(self, form):
        form.save()
        return self.command_response('reload')


class KanbanDescriptionDuplicateModal(Modal):

    def modal_content(self):
        return 'Are you sure you want to duplicate this description?'

    def get_modal_buttons(self):
        return [modal_button_method('Confirm', 'duplicate'), modal_button('Cancel', 'close', 'btn-secondary')]

    def button_duplicate(self, **_kwargs):
        kanban_report_description = get_object_or_404(KanbanReportDescription, id=self.slug['pk'])
        kanban_report_description.pk = None
        kanban_report_description.name = f'Copy of {kanban_report_description.name}'
        kanban_report_description.save()
        return self.command_response('reload')



import calendar
from datetime import datetime, date, timedelta
from calendar import monthrange

from date_offset.date_offset import DateOffset


class VariableDate:
    RANGE_TYPE_TODAY = 1
    RANGE_TYPE_YESTERDAY = 2
    RANGE_TYPE_TOMORROW = 3
    RANGE_TYPE_THIS_WEEK = 4
    RANGE_TYPE_NEXT_WEEK = 5
    RANGE_TYPE_2_WEEKS_TIME = 6
    RANGE_TYPE_3_WEEKS_TIME = 34
    RANGE_TYPE_4_WEEKS_TIME = 35
    RANGE_TYPE_5_WEEKS_TIME = 36
    RANGE_TYPE_6_WEEKS_TIME = 37
    RANGE_TYPE_THIS_MONTH = 7
    RANGE_TYPE_NEXT_MONTH = 8
    RANGE_TYPE_LAST_MONTH = 9
    RANGE_TYPE_NEXT_7_DAYS = 10
    RANGE_TYPE_LAST_7_DAYS = 11
    RANGE_TYPE_NEXT_14_DAYS = 12
    RANGE_TYPE_LAST_14_DAYS = 13
    RANGE_TYPE_NEXT_28_DAYS = 14
    RANGE_TYPE_LAST_28_DAYS = 15
    RANGE_TYPE_THIS_YEAR = 16
    RANGE_TYPE_LAST_YEAR = 17
    RANGE_TYPE_LAST_WEEK = 18
    RANGE_TYPE_LAST_18_MONTHS = 23
    RANGE_TYPE_NEXT_2_MONTH = 24
    RANGE_TYPE_LAST_2_MONTH = 25
    RANGE_TYPE_NEXT_3_MONTH = 26
    RANGE_TYPE_LAST_3_MONTH = 27
    RANGE_TYPE_NEXT_6_MONTH = 28
    RANGE_TYPE_LAST_6_MONTH = 29
    RANGE_TYPE_NEXT_60_DAYS = 30
    RANGE_TYPE_LAST_60_DAYS = 31
    RANGE_TYPE_NEXT_90_DAYS = 32
    RANGE_TYPE_LAST_90_DAYS = 33

    RANGE_TYPE_CHOICES = ((RANGE_TYPE_LAST_6_MONTH, 'Last 6 months'),
                          (RANGE_TYPE_LAST_3_MONTH, 'Last 3 months'),
                          (RANGE_TYPE_LAST_90_DAYS, 'Last 90 days'),
                          (RANGE_TYPE_LAST_2_MONTH, 'Last 2 month'),
                          (RANGE_TYPE_LAST_60_DAYS, 'Last 60 days'),
                          (RANGE_TYPE_LAST_MONTH, 'Last month'),
                          (RANGE_TYPE_LAST_28_DAYS, 'Last 28 days'),
                          (RANGE_TYPE_LAST_14_DAYS, 'Last 14 days'),
                          (RANGE_TYPE_LAST_7_DAYS, 'Last 7 days'),
                          (RANGE_TYPE_LAST_WEEK, 'Last week'),
                          (RANGE_TYPE_YESTERDAY, 'Yesterday'),
                          (RANGE_TYPE_TODAY, 'Today'),
                          (RANGE_TYPE_TOMORROW, 'Tomorrow'),
                          (RANGE_TYPE_THIS_WEEK, 'This week'),
                          (RANGE_TYPE_NEXT_7_DAYS, 'Next 7 days'),
                          (RANGE_TYPE_NEXT_14_DAYS, 'Next 14 days'),
                          (RANGE_TYPE_NEXT_WEEK, 'Next week'),
                          (RANGE_TYPE_2_WEEKS_TIME, '2 weeks time'),
                          (RANGE_TYPE_3_WEEKS_TIME, '3 weeks time'),
                          (RANGE_TYPE_4_WEEKS_TIME, '4 weeks time'),
                          (RANGE_TYPE_5_WEEKS_TIME, '5 weeks time'),
                          (RANGE_TYPE_6_WEEKS_TIME, '6 weeks time'),
                          (RANGE_TYPE_THIS_MONTH, 'This month'),
                          (RANGE_TYPE_NEXT_28_DAYS, 'Next 28 days'),
                          (RANGE_TYPE_NEXT_MONTH, 'Next month'),
                          (RANGE_TYPE_NEXT_2_MONTH, 'Next 2 months'),
                          (RANGE_TYPE_NEXT_60_DAYS, 'Next 60 days'),
                          (RANGE_TYPE_NEXT_3_MONTH, 'Next 3 months'),
                          (RANGE_TYPE_NEXT_90_DAYS, 'Next 90 days'),
                          (RANGE_TYPE_NEXT_6_MONTH, 'Next 6 months'),
                          (RANGE_TYPE_THIS_YEAR, 'This year'),
                          (RANGE_TYPE_LAST_YEAR, 'Last year'),
                          (RANGE_TYPE_LAST_18_MONTHS, 'Last 18 months')
                          )

    def get_variable_dates(self, range_type):
        today = date.today()
        if range_type == self.RANGE_TYPE_TODAY:
            start_date = today
            end_date = today + timedelta(days=1)
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_YESTERDAY:  # Yesterday
            start_date = today - timedelta(days=1)
            end_date = today
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_TOMORROW:  # Tomorrow
            start_date = today + timedelta(days=1)
            end_date = start_date + timedelta(days=1)
            number_of_days = 1
        elif range_type == self.RANGE_TYPE_THIS_WEEK:  # This Week
            start_date = today - timedelta(days=today.weekday())
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_NEXT_WEEK:  # Next Week
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=7)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_LAST_WEEK:  # Last Week
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=14)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_2_WEEKS_TIME:  # 2 Week Time
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=14)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_3_WEEKS_TIME:  # 3 Week Time
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=21)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_4_WEEKS_TIME:  # 4 Week Time
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=28)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_5_WEEKS_TIME:  # 5 Week Time
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=35)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_6_WEEKS_TIME:  # 6 Week Time
            start_date = today - timedelta(days=today.weekday()) + timedelta(days=42)
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_THIS_MONTH:  # This Month
            start_date = date(today.year, today.month, 1)
            number_of_days = monthrange(today.year, today.month)[1]
            end_date = start_date + timedelta(days=number_of_days - 1)
        elif range_type == self.RANGE_TYPE_NEXT_MONTH:  # Next Month
            number_of_days_in_this_month = monthrange(today.year, today.month)[1]
            start_date = (date(today.year, today.month, 1) +
                          timedelta(days=number_of_days_in_this_month))
            number_of_days = monthrange(start_date.year, start_date.month)[1]
            end_date = start_date + timedelta(days=number_of_days)
        elif range_type == self.RANGE_TYPE_LAST_MONTH:  # Last Month
            end_date = date(today.year, today.month, 1) - timedelta(days=1)
            last_month = end_date
            start_date = date(last_month.year, last_month.month, 1)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_7_DAYS:  # Next 7 Days
            start_date = today
            end_date = start_date + timedelta(days=7)
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_LAST_7_DAYS:  # Last 7 Days
            start_date = today - timedelta(days=6)
            end_date = today
            number_of_days = 7
        elif range_type == self.RANGE_TYPE_NEXT_14_DAYS:  # Next 14 Days
            start_date = today
            end_date = start_date + timedelta(days=14)
            number_of_days = 14
        elif range_type == self.RANGE_TYPE_LAST_14_DAYS:  # Next 14 Days
            start_date = today - timedelta(days=14)
            end_date = today
            number_of_days = 14
        elif range_type == 13:  # Last 14 Days
            start_date = today - timedelta(days=14)
            end_date = today
            number_of_days = 14
        elif range_type == self.RANGE_TYPE_NEXT_28_DAYS:  # Next 28 Days
            start_date = today
            end_date = start_date + timedelta(days=28)
            number_of_days = 28
        elif range_type == self.RANGE_TYPE_LAST_28_DAYS:  # Last 28 Days
            start_date = today - timedelta(days=28)
            end_date = today
            number_of_days = 28
        elif range_type == self.RANGE_TYPE_NEXT_60_DAYS:  # Next 60 Days
            start_date = today
            end_date = start_date + timedelta(days=60)
            number_of_days = 60
        elif range_type == self.RANGE_TYPE_LAST_60_DAYS:  # Last 60 Days
            start_date = today - timedelta(days=60)
            end_date = today
            number_of_days = 60
        elif range_type == self.RANGE_TYPE_NEXT_90_DAYS:  # Next 90 Days
            start_date = today
            end_date = start_date + timedelta(days=90)
            number_of_days = 90
        elif range_type == self.RANGE_TYPE_LAST_90_DAYS:  # Last 90 Days
            start_date = today - timedelta(days=90)
            end_date = today
            number_of_days = 90
        elif range_type == self.RANGE_TYPE_THIS_YEAR:  # This Year
            start_date = date(today.year, 1, 1)
            end_date = date(today.year, 12, 31)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_YEAR:  # Last Year
            last_year = today.year - 1
            start_date = date(last_year, 1, 1)
            end_date = date(last_year, 12, 31)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_18_MONTHS:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-18m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_2_MONTH:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-2m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_2_MONTH:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('2m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_3_MONTH:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-3m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_3_MONTH:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('3m', today)
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_LAST_6_MONTH:
            date_offset = DateOffset()
            start_date = date_offset.get_offset('-6m', today)
            end_date = today
            number_of_days = (end_date - start_date).days
        elif range_type == self.RANGE_TYPE_NEXT_6_MONTH:
            date_offset = DateOffset()
            start_date = today
            end_date = date_offset.get_offset('6m', today)
            number_of_days = (end_date - start_date).days
        else:
            assert False, 'unknown date value'

        start_date_and_time = datetime.combine(start_date, datetime.min.time())
        end_date_and_time = datetime.combine(end_date, datetime.max.time())

        return start_date_and_time, end_date_and_time, number_of_days

    def get_variable_date_filter_values(self):
        values = {}
        for choice in self.RANGE_TYPE_CHOICES:
            values['#variable_date:%d' % choice[0]] = choice[1]
        return values

    @staticmethod
    def get_date_filter_years():
        values = {}
        today = date.today()
        start_year = today.year - 10
        end_year = start_year + 15

        for year in range(start_year, end_year):
            values['#year:%d' % year] = year
        return values

    @staticmethod
    def get_date_filter_months():
        values = {}
        for index, month in enumerate(list(calendar.month_name)[1:], 1):
            values['#month:%d' % index] = month
        return values

    @staticmethod
    def get_date_filter_quarters():
        values = {}
        for quarter in range(1, 5):
            values['#quarter:%d' % quarter] = f'Quarter {quarter}'
        for quarter in range(1, 5):
            values['#financial_quarter:%d' % quarter] = f'Financial Quarter {quarter}'
        return values



import copy

from ajax_helpers.mixins import AjaxHelpers
from django.forms import ModelChoiceField
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic import TemplateView
from django_menus.menu import MenuMixin
from django_modals.forms import CrispyForm
from django_modals.modals import ModelFormModal, FormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.select2 import Select2
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.models import Dashboard, DashboardReport, Report
from advanced_report_builder.utils import split_slug
from advanced_report_builder.views.bar_charts import BarChartView
from advanced_report_builder.views.datatables.datatables import TableView
from advanced_report_builder.views.funnel_charts import FunnelChartView
from advanced_report_builder.views.kanban import KanbanView
from advanced_report_builder.views.line_charts import LineChartView
from advanced_report_builder.views.pie_charts import PieChartView
from advanced_report_builder.views.single_values import SingleValueView
from django.conf import settings


class ViewDashboardBase(AjaxHelpers, MenuMixin, TemplateView):
    model = Dashboard
    enable_edit = True
    enable_links = True
    views = {'tablereport': TableView,
             'singlevaluereport': SingleValueView,
             'barchartreport': BarChartView,
             'linechartreport': LineChartView,
             'piechartreport': PieChartView,
             'funnelchartreport': FunnelChartView,
             'kanbanreport': KanbanView,
             }

    custom_views = {}
    views_overrides = {}
    ajax_commands = ['button', 'select2', 'ajax']

    def __init__(self, *args, **kwargs):
        self.dashboard = None
        super().__init__(*args, **kwargs)

    def redirect_url(self):
        """ used if the slug changes"""
        return None

    def dashboard_not_found(self):
        raise Http404

    def dispatch(self, request, *args, **kwargs):
        slug = split_slug(self.kwargs['slug'])
        self.dashboard = self.model.objects.filter(slug=slug['pk']).first()

        if self.dashboard is None:
            self.dashboard = self.model.objects.filter(slug_alias=slug['pk']).first()
            if self.dashboard is None:
                return self.dashboard_not_found()
            else:
                redirect_url = self.redirect_url()
                if redirect_url:
                    return redirect_url

        return super().dispatch(request, *args, **kwargs)

    @staticmethod
    def get_top_report_class(reports):
        reports_len = len(reports)
        spans = {1: ' col-12',
                 2: ' col-12 col-sm-12 col-md-6',
                 3: ' col-12 col-sm-12 col-md-4',
                 4: ' col-12 col-sm-12 col-md-6 col-lg-3',
                 6: ' col-12 col-sm-12 col-md-4',
                 9: ' col-12 col-sm-12 col-md-4'}
        return spans.get(reports_len, ' col-12 col-sm-12 col-md-3')

    def has_dashboard_permission(self):
        """You can over override this to check if the user has permission to view the dashboard.
          If return false 'dashboard_no_permission' will be called"""
        return True

    def dashboard_no_permission(self):
        raise Http404

    def has_report_got_permission(self, report):
        """You can over override this to check if the user has permission to view the report.
          If return false 'report_no_permission' will be called"""
        return True

    def report_no_permission(self, dashboard_report, reports):
        pass

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['dashboard'] = self.dashboard

        if not self.has_dashboard_permission():
            return self.dashboard_no_permission()

        top_reports = []
        reports = []
        for dashboard_report in self.dashboard.dashboardreport_set.all():
            if self.has_report_got_permission(report=dashboard_report.report):
                report_view = self.get_view(report=dashboard_report.report)
                extra_class_name = report_view().get_dashboard_class(report=dashboard_report.report)
                report_data = self.call_view(dashboard_report=dashboard_report, report_view=report_view).rendered_content
                report = {'render': report_data,
                          'name': dashboard_report.report.name,
                          'id': dashboard_report.id,
                          'class': dashboard_report.get_class(extra_class_name=extra_class_name)}
                if dashboard_report.top:
                    top_reports.append(report)
                else:
                    reports.append(report)
            else:
                self.report_no_permission(dashboard_report=dashboard_report, reports=reports)
        context['top_reports'] = top_reports
        context['top_reports_class'] = self.get_top_report_class(top_reports)

        context['reports'] = reports
        context['enable_edit'] = self.enable_edit
        return context

    def get_view(self, report):
        if report.instance_type == 'customreport':
            view_name = report.customreport.view_name
            return self.custom_views.get(view_name)
        elif report.instance_type in self.views_overrides:
            return self.views_overrides.get(report.instance_type)
        return self.views.get(report.instance_type)

    def call_view(self, dashboard_report, report_view=None):
        if report_view is None:
            report_view = self.get_view(report=dashboard_report.report)
        view_kwargs = copy.deepcopy(self.kwargs)
        view_kwargs['report'] = dashboard_report.report
        view_kwargs['dashboard_report'] = dashboard_report
        view_kwargs['enable_edit'] = self.enable_edit
        view_kwargs['enable_links'] = self.enable_links
        return report_view.as_view()(self.request, *self.args, **view_kwargs)

    def post(self, request, *args, **kwargs):
        table_id = request.POST.get('table_id')
        if table_id:   # must be a datatable:
            dashboard_report_id = table_id.split('_')[1]
            if dashboard_report_id:
                dashboard_report = self.dashboard.dashboardreport_set.filter(id=dashboard_report_id).first()
                return self.call_view(dashboard_report=dashboard_report)

        return super().post(request, *args, **kwargs)

    def ajax_change_placement(self, **kwargs):

        ids = kwargs['ids']
        dashboard_reports = DashboardReport.objects.filter(id__in=ids)
        obj_dict = dict([(obj.id, obj) for obj in dashboard_reports])
        sorted_objects = [obj_dict[_id] for _id in ids]

        for index, dashboard_report in enumerate(sorted_objects):
            if dashboard_report.order != index:
                dashboard_report.order = index
                dashboard_report.save()

        return self.command_response()


class DashboardModal(ModelFormModal):
    model = Dashboard
    form_fields = ['name',
                   'display_option']
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF

    def post_save(self, created, form):
        if created:
            url_name = getattr(settings, 'REPORT_BUILDER_DASHBOARD_URL_NAME', '')
            if url_name:
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.add_command('redirect', url=url)


class DashboardReportModal(ModelFormModal):
    model = DashboardReport
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF

    form_fields = ['name_override',
                   'top',
                   'display_option']
    widgets = {'top': Toggle(attrs={'data-onstyle': 'success', 'data-on': 'YES', 'data-off': 'NO'})}

    @staticmethod
    def form_setup(form, *_args, **_kwargs):
        form.add_trigger('top', 'onchange', [
            {'selector': '#div_id_display_option', 'values': {'checked': 'hide'}, 'default': 'show'},
        ])


class DashboardAddReportForm(CrispyForm):
    report = ModelChoiceField(queryset=Report.objects.all(), widget=Select2)


class DashboardAddReportModal(FormModal):
    form_class = DashboardAddReportForm
    modal_title = 'Add Report'

    def form_valid(self, form):
        dashboard = get_object_or_404(Dashboard, id=self.slug['pk'])
        dashboard_report = DashboardReport(dashboard=dashboard)
        dashboard_report.report = form.cleaned_data['report']
        dashboard_report.save()
        return self.command_response('reload')



from django.conf import settings
from django.urls import reverse
from django_modals.modals import ModelFormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF

from advanced_report_builder.models import Target


class TargetModal(ModelFormModal):
    model = Target
    form_fields = ['name',
                   'target_type',
                   'colour',
                   'default_value']

    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF

    def post_save(self, created, form):
        if created:
            url_name = getattr(settings, 'REPORT_BUILDER_TARGET_URL_NAME', '')
            if url_name:
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.add_command('redirect', url=url)



# Generated by Django 3.2.7 on 2022-02-07 15:37

from django.db import migrations, models
import django.db.models.deletion
import time_stamped_model.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Dashboard',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('slug', models.SlugField(unique=True)),
                ('slug_alias', models.SlugField(blank=True, null=True)),
                ('name', models.CharField(max_length=200)),
                ('display_option', models.PositiveIntegerField(choices=[(1, '1 Report per Row'), (2, '2 Reports per Row'), (3, '3 Reports per Row'), (4, '4 Reports per Row')], default=2)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('slug', models.SlugField(unique=True)),
                ('slug_alias', models.SlugField(blank=True, null=True)),
                ('instance_type', models.CharField(max_length=255, null=True)),
                ('notes', models.TextField(blank=True, null=True)),
                ('version', models.PositiveSmallIntegerField(default=0)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='ReportTag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=128, unique=True)),
                ('slug', models.SlugField(unique=True)),
                ('order', models.PositiveSmallIntegerField()),
            ],
            options={
                'ordering': ('order',),
            },
        ),
        migrations.CreateModel(
            name='BarChartReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('axis_scale', models.PositiveSmallIntegerField(choices=[(1, 'Year'), (2, 'Quarter'), (3, 'Month'), (4, 'Week'), (5, 'Day')])),
                ('date_field', models.CharField(max_length=200)),
                ('axis_value_type', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Max'), (3, 'Min'), (4, 'Count'), (5, 'Avg')], default=4, null=True)),
                ('fields', models.JSONField(blank=True, null=True)),
                ('x_label', models.CharField(blank=True, max_length=200, null=True)),
                ('y_label', models.CharField(blank=True, max_length=200, null=True)),
                ('bar_chart_orientation', models.PositiveSmallIntegerField(choices=[(1, 'Vertical'), (2, 'Horizontal')], default=1)),
                ('stacked', models.BooleanField(default=False)),
                ('show_totals', models.BooleanField(default=False)),
                ('show_blank_dates', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='FunnelChartReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('axis_value_type', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Max'), (3, 'Min'), (4, 'Count'), (5, 'Avg')], default=4, null=True)),
                ('fields', models.JSONField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='KanbanReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='LineChartReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('axis_scale', models.PositiveSmallIntegerField(choices=[(1, 'Year'), (2, 'Quarter'), (3, 'Month'), (4, 'Week'), (5, 'Day')])),
                ('date_field', models.CharField(max_length=200)),
                ('axis_value_type', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Max'), (3, 'Min'), (4, 'Count'), (5, 'Avg')], default=4, null=True)),
                ('fields', models.JSONField(blank=True, null=True)),
                ('x_label', models.CharField(blank=True, max_length=200, null=True)),
                ('y_label', models.CharField(blank=True, max_length=200, null=True)),
                ('show_totals', models.BooleanField(default=False)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='PieChartReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('axis_value_type', models.PositiveSmallIntegerField(blank=True, choices=[(1, 'Sum'), (2, 'Max'), (3, 'Min'), (4, 'Count'), (5, 'Avg')], default=4, null=True)),
                ('fields', models.JSONField(blank=True, null=True)),
                ('style', models.PositiveSmallIntegerField(choices=[(1, 'Pie'), (2, 'Doughnut')], default=1)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='SingleValueReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('tile_colour', models.CharField(blank=True, max_length=10, null=True)),
                ('field', models.CharField(blank=True, max_length=200, null=True)),
                ('numerator', models.CharField(blank=True, max_length=200, null=True)),
                ('single_value_type', models.PositiveSmallIntegerField(choices=[(1, 'Count'), (2, 'Sum'), (3, 'Count & Sum'), (4, 'Percent'), (5, 'Percent from Count'), (6, 'Average')], default=1)),
                ('prefix', models.CharField(blank=True, max_length=64, null=True)),
                ('decimal_places', models.IntegerField(default=0)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='TableReport',
            fields=[
                ('report_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='advanced_report_builder.report')),
                ('table_fields', models.JSONField(blank=True, null=True)),
                ('has_clickable_rows', models.BooleanField(default=False)),
                ('link_field', models.CharField(blank=True, max_length=200, null=True)),
                ('pivot_fields', models.JSONField(blank=True, null=True)),
                ('page_length', models.PositiveSmallIntegerField(choices=[(10, '10'), (25, '25'), (50, '50'), (100, '100'), (150, '150'), (200, '200')], default=100)),
            ],
            options={
                'abstract': False,
            },
            bases=('advanced_report_builder.report',),
        ),
        migrations.CreateModel(
            name='ReportType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('slug', models.SlugField(unique=True)),
                ('report_builder_class_name', models.CharField(max_length=200)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.contenttype')),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.AddField(
            model_name='report',
            name='report_tags',
            field=models.ManyToManyField(blank=True, to='advanced_report_builder.ReportTag'),
        ),
        migrations.AddField(
            model_name='report',
            name='report_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype'),
        ),
        migrations.CreateModel(
            name='DashboardReport',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('order', models.PositiveSmallIntegerField()),
                ('top', models.BooleanField(default=False)),
                ('name_override', models.CharField(blank=True, max_length=200, null=True)),
                ('display_option', models.PositiveIntegerField(choices=[(0, 'None/Default'), (1, '1 Report per Row'), (2, '2 Reports per Row'), (3, '3 Reports per Row'), (4, '4 Reports per Row')], default=0)),
                ('dashboard', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.dashboard')),
                ('report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.report')),
            ],
            options={
                'ordering': ['order'],
            },
        ),
        migrations.CreateModel(
            name='ReportQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.TextField(default='Standard')),
                ('query', models.JSONField(blank=True, null=True)),
                ('extra_query', models.JSONField(blank=True, null=True)),
                ('report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.report')),
            ],
            options={
                'verbose_name_plural': 'Report queries',
                'ordering': ['name'],
                'unique_together': {('name', 'report')},
            },
        ),
        migrations.CreateModel(
            name='KanbanReportLane',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', time_stamped_model.models.CreationDateTimeField(auto_now_add=True)),
                ('modified', time_stamped_model.models.ModificationDateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200)),
                ('order', models.PositiveSmallIntegerField()),
                ('heading_field', models.CharField(blank=True, max_length=200, null=True)),
                ('order_by_field', models.CharField(blank=True, max_length=200, null=True)),
                ('link_field', models.CharField(blank=True, max_length=200, null=True)),
                ('order_by_ascending', models.BooleanField(default=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('query_data', models.JSONField(blank=True, null=True)),
                ('report_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='advanced_report_builder.reporttype')),
                ('kanban_report', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='advanced_report_builder.kanbanreport')),
            ],
            options={
                'ordering': ('name',),
            },
        ),
    ]



# Generated by Django 3.2.7 on 2023-06-20 14:49

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0010_auto_20230428_2033'),
    ]

    operations = [
        migrations.AddField(
            model_name='barchartreport',
            name='date_field_type',
            field=models.PositiveSmallIntegerField(choices=[(1, 'Single'), (2, 'Range')], default=1),
        ),
        migrations.AddField(
            model_name='barchartreport',
            name='end_date_field',
            field=models.CharField(blank=True, max_length=200, null=True),
        ),
    ]



import base64
import json

from crispy_forms.bootstrap import StrictButton
from crispy_forms.layout import Div, HTML
from django.conf import settings
from django.forms import CharField, ChoiceField, BooleanField, IntegerField
from django.urls import reverse
from django_datatables.columns import MenuColumn
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import HtmlMenu, MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.forms import ModelCrispyForm
from django_modals.modals import FormModal, ModelFormModal
from django_modals.processes import PROCESS_EDIT_DELETE, PERMISSION_OFF
from django_modals.widgets.select2 import Select2Multiple
from django_modals.widgets.widgets import Toggle

from advanced_report_builder.globals import DATE_FIELDS, NUMBER_FIELDS, ANNOTATION_VALUE_CHOICES, ANNOTATIONS_CHOICES, \
    DATE_FORMAT_TYPES, CURRENCY_COLUMNS, LINK_COLUMNS
from advanced_report_builder.models import TableReport, ReportQuery, ReportType
from advanced_report_builder.toggle import RBToggle
from advanced_report_builder.utils import split_attr, get_field_details, encode_attribute, decode_attribute, \
    get_report_builder_class
from advanced_report_builder.views.charts_base import ChartBaseFieldForm
from advanced_report_builder.views.modals_base import QueryBuilderModalBaseMixin, QueryBuilderModalBase


class TableModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/datatables/modal.html'
    size = 'xl'
    model = TableReport
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    ajax_commands = ['datatable', 'button']

    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name',
                   'notes',
                   ('has_clickable_rows', {'widget': Toggle(attrs={'data-onstyle': 'success',
                                                                   'data-on': 'YES',
                                                                   'data-off': 'NO'})}),
                   'link_field',
                   'order_by_field',
                   ('order_by_ascending', {'widget': Toggle(attrs={'data-onstyle': 'success',
                                                                   'data-on': 'YES',
                                                                   'data-off': 'NO'})}),

                   'page_length',
                   'report_type',
                   'report_tags',
                   'table_fields',
                   'pivot_fields',
                   ]

    def form_setup(self, form, *_args, **_kwargs):
        url = reverse('advanced_report_builder:table_field_modal',
                      kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        pivot_url = reverse('advanced_report_builder:table_pivot_modal',
                            kwargs={'slug': 'selector-99999-data-FIELD_INFO-report_type_id-REPORT_TYPE_ID'})

        form.add_trigger('has_clickable_rows', 'onchange', [
            {'selector': '#div_id_link_field', 'values': {'checked': 'show'}, 'default': 'hide'},
        ])

        form.fields['notes'].widget.attrs['rows'] = 3

        if 'data' in _kwargs:
            link_field = _kwargs['data'].get('link_field')
            order_by_field = _kwargs['data'].get('order_by_field')
            report_type_id = _kwargs['data'].get('report_type')
            report_type = ReportType.objects.filter(id=report_type_id).first()  # can be None
        else:
            link_field = form.instance.link_field
            order_by_field = form.instance.order_by_field
            report_type = form.instance.report_type

        self.setup_field(field_type='link',
                         form=form,
                         field_name='link_field',
                         selected_field_id=link_field,
                         report_type=report_type)

        self.setup_field(field_type='order',
                         form=form,
                         field_name='order_by_field',
                         selected_field_id=order_by_field,
                         report_type=report_type)

        fields = ['name',
                  'notes',
                  'report_type',
                  'report_tags',
                  FieldEx('has_clickable_rows', template='django_modals/fields/label_checkbox.html'),
                  'link_field',
                  'order_by_field',
                  FieldEx('order_by_ascending', template='django_modals/fields/label_checkbox.html'),
                  FieldEx('page_length', template='django_modals/fields/label_checkbox.html'),
                  FieldEx('table_fields',
                          template='advanced_report_builder/select_column.html',
                          extra_context={'select_column_url': url,
                                         'command_prefix': ''}),
                  FieldEx('pivot_fields',
                          template='advanced_report_builder/datatables/select_pivot.html',
                          extra_context={'select_column_url': pivot_url}),
                  ]

        if self.object.id:

            add_query_js = 'django_modal.process_commands_lock([{"function": "post_modal", ' \
                            '"button": {"button": "add_query"}}])'

            description_add_menu_items = [

                MenuItem(add_query_js.replace('"', "&quot;"),
                         menu_display='Add Query Version',
                         css_classes='btn btn-primary',
                         font_awesome='fas fa-pencil',
                         link_type=MenuItem.HREF)]

            menu = HtmlMenu(self.request,
                            'advanced_report_builder/datatables/onclick_menu.html').add_items(
                            *description_add_menu_items)

            fields.append(Div(HTML(menu.render()), css_class='form-buttons'))
            edit_query_js = 'django_modal.process_commands_lock([{"function": "post_modal", ' \
                            '"button": {"button": "edit_query", "query_id": $(this).closest(\'tr\').attr(\'id\')}}])'

            duplicate_query_js = 'django_modal.process_commands_lock([{"function": "post_modal", ' \
                                 '"button": {"button": "duplicate_query", ' \
                                 '"query_id": $(this).closest(\'tr\').attr(\'id\')}}])'

            description_edit_menu_items = [
                MenuItem(duplicate_query_js.replace('"', "&quot;"),
                         menu_display='Duplicate',
                         css_classes='btn btn-sm btn-outline-dark',
                         font_awesome='fas fa-clone',
                         link_type=MenuItem.HREF),

                MenuItem(edit_query_js.replace('"', "&quot;"),
                         menu_display='Edit',
                         css_classes='btn btn-sm btn-outline-dark',
                         font_awesome='fas fa-pencil',
                         link_type=MenuItem.HREF)]

            form.fields['queries'] = CharField(
                required=False,
                label='Queries Versions',
                widget=DataTableReorderWidget(
                    model=ReportQuery,
                    order_field='order',
                    fields=['_.index',
                            '.id',
                            'name',

                            MenuColumn(column_name='menu', field='id',
                                       column_defs={'orderable': False, 'className': 'dt-right'},
                                       menu=HtmlMenu(self.request,
                                                     'advanced_report_builder/datatables/onclick_menu.html').add_items(
                                           *description_edit_menu_items)),
                            ],
                    attrs={'filter': {'report__id': self.object.id}}))
            fields.append('queries')
        return fields

    def form_valid(self, form):
        org_id = self.object.pk if hasattr(self, 'object') else None
        save_function = getattr(form, 'save', None)
        if save_function:
            save_function()
        self.post_save(created=org_id is None, form=form)
        if not self.response_commands:
            self.add_command('reload')
        return self.command_response()

    def post_save(self, created, form):
        if created:
            self.modal_redirect(self.request.resolver_match.view_name, slug=f'pk-{self.object.id}-new-True')
        else:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if url_name and self.slug.get('new'):
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.command_response('redirect', url=url)

    def ajax_get_fields(self, **kwargs):
        report_type_id = kwargs['report_type']
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=report_type_id)
        fields = []
        tables = []
        pivot_fields = []
        self._get_fields(base_model=base_model,
                         fields=fields,
                         tables=tables,
                         report_builder_class=report_builder_class,
                         pivot_fields=pivot_fields)

        self.add_command('report_fields', data=json.dumps({'fields': fields, 'tables': tables}))
        self.add_command('report_pivots', data=json.dumps({'pivot_fields': pivot_fields}))

        return self.command_response()

    def select2_link_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='link',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def select2_order_by_field(self, **kwargs):
        return self.get_fields_for_select2(field_type='order',
                                           report_type=kwargs['report_type'],
                                           search_string=kwargs.get('search'))

    def datatable_sort(self, **kwargs):

        form = self.get_form()
        widget = form.fields[kwargs['table_id'][3:]].widget
        _model = widget.attrs['table_model']
        current_sort = dict(_model.objects.filter(report=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = _model.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('')

    def button_duplicate_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_query = ReportQuery.objects.get(pk=query_id)
        report_query.slug = None
        report_query.pk = None
        report_query.order = None
        report_query.save()

        report_type = _kwargs['report_type']
        url = reverse('advanced_report_builder:table_query_modal',
                      kwargs={'slug': f'pk-{report_query.id}-report_type-{report_type}'})
        return self.command_response('show_modal', modal=url)

    def button_add_query(self, **_kwargs):
        report_type = _kwargs['report_type']
        url = reverse('advanced_report_builder:table_query_modal',
                      kwargs={'slug': f'report_id-{self.object.id}-report_type-{report_type}'})
        return self.command_response('show_modal', modal=url)

    def button_edit_query(self, **_kwargs):
        query_id = _kwargs['query_id'][1:]
        report_type = _kwargs['report_type']
        url = reverse('advanced_report_builder:table_query_modal',
                      kwargs={'slug': f'pk-{query_id}-report_type-{report_type}'})
        return self.command_response('show_modal', modal=url)


class TableFieldForm(ChartBaseFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)

    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        if self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            return StrictButton(button_text, onclick=f'save_modal_{self.form_id}()', css_class=css_class, **kwargs)
        else:
            return super().submit_button(css_class, button_text, **kwargs)

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        report_type, base_model = self.get_report_type_details()

        self.fields['title'] = CharField(initial=data['title'])

        data_attr = split_attr(data)
        if self.django_field is not None and isinstance(self.django_field, DATE_FIELDS):
            self.fields['annotations_value'] = ChoiceField(choices=[(0, '-----')] + ANNOTATION_VALUE_CHOICES,
                                                           required=False)
            if 'annotations_value' in data_attr:
                self.fields['annotations_value'].initial = data_attr['annotations_value']
            self.fields['date_format'] = ChoiceField(choices=[(0, '-----')] + DATE_FORMAT_TYPES, required=False)
            if 'date_format' in data_attr:
                self.fields['date_format'].initial = data_attr['date_format']
        elif self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            self.fields['annotations_type'] = ChoiceField(choices=[(0, '-----')] + ANNOTATIONS_CHOICES,
                                                          required=False)
            if 'annotations_type' in data_attr:
                self.fields['annotations_type'].initial = data_attr['annotations_type']
            self.fields['show_table_totals'] = BooleanField(required=False,
                                                            widget=RBToggle(),
                                                            label='Show totals')
            if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
                self.fields['show_table_totals'].initial = True
            self.fields['decimal_places'] = IntegerField()
            self.fields['decimal_places'].initial = int(data_attr.get('decimal_places', 0))
            self.fields['has_filter'] = BooleanField(required=False, widget=RBToggle())

            self.fields['filter'] = CharField(required=False)

            if data_attr.get('has_filter') == '1':
                self.fields['has_filter'].initial = True
                if 'filter' in data_attr:
                    self.fields['filter'].initial = decode_attribute(data_attr['filter'])

            self.fields['multiple_columns'] = BooleanField(required=False, widget=RBToggle())

            report_builder_class = get_report_builder_class(model=base_model,
                                                            report_type=report_type)
            fields = []
            self._get_query_builder_foreign_key_fields(base_model=base_model,
                                                       report_builder_class=report_builder_class,
                                                       fields=fields)

            self.fields['multiple_column_field'] = ChoiceField(choices=fields, required=False)

            if data_attr.get('multiple_columns') == '1':
                self.fields['multiple_columns'].initial = True
                self.fields['multiple_column_field'].initial = data_attr.get('multiple_column_field')
        elif isinstance(self.col_type_override, CURRENCY_COLUMNS):
            self.fields['show_table_totals'] = BooleanField(required=False,
                                                            widget=RBToggle(),
                                                            label='Show totals')
            if 'show_totals' in data_attr and data_attr['show_totals'] == '1':
                self.fields['show_table_totals'].initial = True
        elif isinstance(self.col_type_override, LINK_COLUMNS):
            self.fields['link_html'] = CharField(required=False)
            if 'link_html' in data_attr:
                self.fields['link_html'].initial = decode_attribute(data_attr['link_html'])
            self.fields['link_css'] = CharField(required=False)
            if 'link_css' in data_attr:
                self.fields['link_css'].initial = decode_attribute(data_attr['link_css'])
            self.fields['is_icon'] = BooleanField(required=False, widget=RBToggle())
            if 'is_icon' in data_attr and data_attr['is_icon'] == '1':
                self.fields['is_icon'].initial = True
        else:
            self.fields['annotation_label'] = BooleanField(required=False, widget=RBToggle())
            if 'annotation_label' in data_attr and data_attr['annotation_label'] == '1':
                self.fields['annotation_label'].initial = True

        super().setup_modal(*args, **kwargs)

    def get_additional_attributes(self):
        attributes = []
        self.get_report_type_details()
        if self.django_field is not None and isinstance(self.django_field, DATE_FIELDS):
            if self.cleaned_data['annotations_value']:
                attributes.append(f'annotations_value-{self.cleaned_data["annotations_value"]}')
            if self.cleaned_data['date_format']:
                attributes.append(f'date_format-{self.cleaned_data["date_format"]}')
        elif self.django_field is not None and isinstance(self.django_field, NUMBER_FIELDS):
            if int(self.cleaned_data['annotations_type']) != 0:
                attributes.append(f'annotations_type-{self.cleaned_data["annotations_type"]}')
            if self.cleaned_data['show_table_totals']:
                attributes.append('show_totals-1')
            if self.cleaned_data['decimal_places'] > 0:
                attributes.append(f'decimal_places-{self.cleaned_data["decimal_places"]}')
            if self.cleaned_data['has_filter']:
                attributes.append(f'has_filter-1')
                if self.cleaned_data['filter']:
                    b64_filter = encode_attribute(self.cleaned_data['filter'])
                    attributes.append(f'filter-{b64_filter}')
                if self.cleaned_data['multiple_columns']:
                    attributes.append('multiple_columns-1')
                    attributes.append(f'multiple_column_field-{self.cleaned_data["multiple_column_field"]}')
        elif isinstance(self.col_type_override, CURRENCY_COLUMNS):
            if self.cleaned_data['show_table_totals']:
                attributes.append('show_totals-1')
        elif isinstance(self.col_type_override, LINK_COLUMNS):
            if self.cleaned_data['link_css']:
                b64_link_css = encode_attribute(self.cleaned_data['link_css'])
                attributes.append(f'link_css-{b64_link_css}')
            if self.cleaned_data['link_html']:
                b64_link_html = encode_attribute(self.cleaned_data['link_html'])
                attributes.append(f'link_html-{b64_link_html}')
            if self.cleaned_data['is_icon'] and self.cleaned_data["is_icon"]:
                attributes.append('is_icon-1')
        else:
            if self.cleaned_data['annotation_label'] and self.cleaned_data["annotation_label"]:
                attributes.append('annotation_label-1')

        if attributes:
            return '-'.join(attributes)
        return None


class TableFieldModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = TableFieldForm
    size = 'xl'
    template_name = 'advanced_report_builder/datatables/fields/modal.html'
    no_header_x = True
    helper_class = HorizontalNoEnterHelper

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        _attr = form.get_additional_attributes()
        self.add_command({'function': 'set_attr',
                          'selector': f'#{selector}',
                          'attr': 'data-attr',
                          'val': _attr})

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})
        self.add_command({'function': 'update_selection'})
        return self.command_response('close')

    def form_setup(self, form, *_args, **_kwargs):

        data = json.loads(base64.b64decode(self.slug['data']))
        report_builder_class, base_model = self.get_report_builder_class(report_type_id=self.slug['report_type_id'])
        django_field, col_type_override, _, _ = get_field_details(base_model=base_model,
                                                                  field=data['field'],
                                                                  report_builder_class=report_builder_class)
        if django_field is not None and isinstance(django_field, NUMBER_FIELDS):
            form.add_trigger('annotations_type', 'onchange', [
                {'selector': '#annotations_fields_div', 'values': {'': 'hide'}, 'default': 'show'}])

            form.add_trigger('has_filter', 'onchange', [
                {'selector': '#filter_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'}])

            form.add_trigger('multiple_columns', 'onchange', [
                {'selector': '#multiple_columns_fields_div', 'values': {'checked': 'show'}, 'default': 'hide'},
            ])

            return ['title',
                    'annotations_type',
                    'show_table_totals',
                    'decimal_places',
                    Div(FieldEx('has_filter',
                                template='django_modals/fields/label_checkbox.html',
                                field_class='col-6 input-group-sm'),
                        Div(
                            FieldEx('filter',
                                    template='advanced_report_builder/datatables/fields/single_query_builder.html'),
                            FieldEx('multiple_columns',
                                    template='django_modals/fields/label_checkbox.html',
                                    field_class='col-6 input-group-sm'),
                            Div(
                                FieldEx('multiple_column_field'),
                                css_id='multiple_columns_fields_div'),
                            css_id='filter_fields_div'),
                        css_id='annotations_fields_div')
                    ]
        elif isinstance(col_type_override, CURRENCY_COLUMNS):
            return ['title',
                    'show_table_totals']
        elif isinstance(col_type_override, LINK_COLUMNS):
            return ['title',
                    'link_css',
                    'link_html',
                    'is_icon']

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type_id']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))


class TablePivotForm(ChartBaseFieldForm):
    cancel_class = 'btn-secondary modal-cancel'

    def setup_modal(self, *args, **kwargs):
        data = json.loads(base64.b64decode(self.slug['data']))
        self.fields['title'] = CharField(initial=data['title'])
        super().setup_modal(*args, **kwargs)

    def cancel_button(self, css_class=cancel_class, **kwargs):
        commands = [{'function': 'close'}]
        return self.button('Cancel', commands, css_class, **kwargs)


class TablePivotModal(QueryBuilderModalBaseMixin, FormModal):
    form_class = TablePivotForm
    size = 'xl'
    no_header_x = True

    @property
    def modal_title(self):
        data = json.loads(base64.b64decode(self.slug['data']))
        return f'Edit {data["title"]}'

    def form_valid(self, form):
        selector = self.slug['selector']

        self.add_command({'function': 'html', 'selector': f'#{selector} span', 'html': form.cleaned_data['title']})

        self.add_command({'function': 'update_pivot_selection'})
        return self.command_response('close')


class TableQueryForm(ModelCrispyForm):
    cancel_class = 'btn-secondary modal-cancel'

    class Meta:
        model = ReportQuery
        fields = ['name',
                  'query']

    def submit_button(self, css_class='btn-success modal-submit', button_text='Submit', **kwargs):
        return StrictButton(button_text, onclick=f'save_modal_{self.form_id}()', css_class=css_class, **kwargs)


class TableQueryModal(QueryBuilderModalBaseMixin, ModelFormModal):
    model = ReportQuery
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    form_class = TableQueryForm
    helper_class = HorizontalNoEnterHelper

    template_name = 'advanced_report_builder/datatables/query_modal.html'
    no_header_x = True

    def post_save(self, created, form):
        self.add_command({'function': 'save_query_builder_id_query'})
        return self.command_response('close')

    def form_setup(self, form, *_args, **_kwargs):
        fields = ['name',
                  FieldEx('query', template='advanced_report_builder/query_builder.html')]
        return fields

    def ajax_get_query_builder_fields(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']

        report_type_id = self.slug['report_type']
        query_builder_filters = self.get_query_builder_report_type_field(report_type_id=report_type_id)

        return self.command_response(f'query_builder_{field_auto_id}', data=json.dumps(query_builder_filters))
